<?php

namespace Drupal\image_scale_and_crop_without_upscale\Plugin\ImageEffect;

use Drupal\Core\Image\ImageInterface;
use Drupal\image\Plugin\ImageEffect\ScaleAndCropImageEffect;

/**
 * Scales and crops an image resource without upscaling.
 *
 * Differs from the main ScaleAndCrop effect in that it will not upscale the
 * original image in order to "fit" the desired image dimensions. If either or
 * both of the original image dimensions are too small to fit the desired
 * dimensions, then instead of scaling the image up, the desired dimensions are
 * scaled down in order to fix the original image. As a result, the aspect
 * ratio of the desired dimensions is respected, but the overall size of the
 * image may be smaller.
 *
 * @ImageEffect(
 *   id = "image_scale_and_crop_without_upscale",
 *   label = @Translation("Scale and crop (without upscale)"),
 *   description = @Translation("Scale and crop will maintain the aspect-ratio of the original image, then crop the larger dimension. No upscaling is performed.")
 * )
 */
class ScaleAndCropWithoutUpscaleImageEffect extends ScaleAndCropImageEffect {

  /**
   * {@inheritdoc}
   */
  public function applyEffect(ImageInterface $image) {
    $targetWidth = $this->configuration['width'];
    $targetHeight = $this->configuration['height'];

    // If there's not enough image to completely fill the target dimensions,
    // then shrink the target dimensions so that it does fit. This prevents
    // us from pointlessly upscaling the image (and this reducing quality).
    // There's three cases to handle. The first if both the source width and
    // height are smaller than the target width and height.
    if ($image->getWidth() <= $targetWidth && $image->getHeight() <= $targetHeight) {
      // If the source ratio is already the same as the target ratio, then
      // there's nothing for us to do at all here, the image is already good
      // as it is. Since we have access to the aspect ratio as a float,
      // the calculated source and target ratios may not be precisly the same
      // ratio, but we just want to check that they're close enough.
      $targetRatio = round($targetWidth / $targetHeight, 2, PHP_ROUND_HALF_UP);
      $sourceRatio = round($image->getWidth() / $image->getHeight(), 2, PHP_ROUND_HALF_UP);
      if ($sourceRatio === $targetRatio) {
        return TRUE;
      }

      // Depending on how the A/Rs compare between the source and target,
      // figure out if we need to adjust the target's width or height.
      if ($targetRatio > $sourceRatio) {
        $this->adjustTargetDimensionsToMatchSourceWidth($image);
      }
      else {
        $this->adjustTargetDimensionsToMatchSourceHeight($image);
      }
    }
    // The second case is if just the image width is smaller than the target
    // width.
    elseif ($image->getWidth() <= $targetWidth) {
      $this->adjustTargetDimensionsToMatchSourceWidth($image);
    }
    // And the final case is if just the image height is smaller than the
    // target height.
    elseif ($image->getHeight() <= $targetHeight) {
      $this->adjustTargetDimensionsToMatchSourceHeight($image);
    }

    return parent::applyEffect($image);
  }

  /**
   * Scale target dimensions down to match the height of the source.
   *
   * The aspect ratio of the target dimensions is preserved.
   *
   * @param \Drupal\Core\Image\ImageInterface $image
   *   The source image.
   */
  private function adjustTargetDimensionsToMatchSourceHeight(ImageInterface $image) {
    $targetRatio = $this->configuration['width'] / $this->configuration['height'];
    $this->configuration['height'] = $image->getHeight();
    $this->configuration['width'] = round($this->configuration['height'] * $targetRatio, 0, PHP_ROUND_HALF_UP);
  }

  /**
   * Scale target dimensions down to match the width of the source.
   *
   * The aspect ratio of the target dimensions is preserved.
   *
   * @param \Drupal\Core\Image\ImageInterface $image
   *   The source image.
   */
  private function adjustTargetDimensionsToMatchSourceWidth(ImageInterface $image) {
    $targetRatio = $this->configuration['width'] / $this->configuration['height'];
    $this->configuration['width'] = $image->getWidth();
    $this->configuration['height'] = round($this->configuration['width'] / $targetRatio, 0, PHP_ROUND_HALF_UP);
  }

}
